# coding: utf-8
__all__ = ['querymanager', 'Query', 'load_query']

# Standard Python libraries
from typing import Optional


# Relative imports
from .Query import Query

# Initialize a ModuleManager for the query styles
from ..tools import ModuleManager
querymanager = ModuleManager('Query')

# Add the modular Query styles (future)
querymanager.import_style('bool_match', '.BoolMatchQuery', __name__)
querymanager.import_style('str_contains', '.StrContainsQuery', __name__)
querymanager.import_style('str_match', '.StrMatchQuery', __name__)
querymanager.import_style('list_contains', '.ListContainsQuery', __name__)
querymanager.import_style('int_match', '.IntMatchQuery', __name__)
querymanager.import_style('float_match', '.FloatMatchQuery', __name__)
querymanager.import_style('date_match', '.DateMatchQuery', __name__)
querymanager.import_style('month_match', '.MonthMatchQuery', __name__)
querymanager.import_style('dummy', '.DummyQuery', __name__)

def load_query(style: str,
               name: Optional[str] = None,
               parent: Optional[str] = None,
               path: Optional[str] = None,
               description: Optional[str] = None,
               **kwargs) -> Query:
    """
    Loads a Query subclass associated with a given query style.

    Parameters
    ----------
    style : str
        The query style.
    name : str or None, optional
        The metadata key associated with the data field.  Must be set
        to use the pandas query method.
    parent : str or None, optional
        Allows for the pandas query operations to work on embedded
        metadata dicts.  If given, the pandas query method will check the
        value of metadata[parent][name].
    path : str or None, optional
        The record data path to the data field.  Levels are delimited by
        periods.  Must be given to use the mongo query method.
    description : str, optional
        Description of the query operation, i.e. what it is searching.
    **kwargs : any, optional
        Any additional style-specific keyword parameters.
    """
    return querymanager.init(style, name=name, parent=parent, path=path,
                             description=description, **kwargs)


# coding: utf-8
__all__ = sorted(['aslist', 'iaslist', 'screen_input', 'dict_insert',
                  'ModuleManager'])

# Relative imports
from cdcs import aslist, iaslist
from .screen_input import screen_input
from .dict_insert import dict_insert
from .ModuleManager import ModuleManager


from typing import Optional, Union, Any

import numpy as np
import numpy.typing as npt

from . import Value
from .. import unitconvert as uc

class IntArrayValue(Value):
    
    def __init__(self,
                 name: str,
                 record,
                 defaultvalue: Optional[Any] = None,
                 valuerequired: bool = False,
                 allowedvalues: Optional[tuple] = None,
                 metadatakey: Union[str, bool, None] = False,
                 metadataparent: Optional[str] = None,
                 modelpath: Optional[str] = None,
                 description: Optional[str] = None):
        """
        Initialize an FloatArrayValue object for managing arrays of floats.

        Parameters
        ----------
        name : str
            The name of the parameter.  This should correspond to the name of
            the associated class attribute.
        record : Record
            The Record object that the Parameter is used with.
        defaultvalue : any or None, optional
            The default value to use for the property.  The default value of
            None indicates that there is no default value.
        valuerequired: bool
            Indicates if a value must be given for the property.  If True, then
            checks will be performed that a value is assigned to the property.
        allowedvalues : tuple or None, optional
            A list/tuple of values that the parameter is restricted to have.
            Setting this to None (default) indicates any value is allowed.
        metadatakey: str, bool or None, optional
            The key name to use for the property when constructing the record
            metadata dict.  If set to None (default) then name will be used for
            metadatakey.  If set to False then the parameter will not be
            included in the metadata dict.
        metadataparent: str or None, optional
            If given, then this indicates that the metadatakey is actually an
            element of a dict in metadata with this name.  This allows for limited
            support for metadata having embedded dicts.
        modelpath: str, optional
            The period-delimited path after the record root element for
            where the parameter will be found in the built data model.  If set
            to None (default) then name will be used for modelpath.
        description: str or None, optional
            A short description for the value.  If not given, then the record name
            will be used.
        """
        # Require metadatakey to be explicitly given
        if metadatakey is None:
            metadatakey = False
        
        # Check that allowedvalues is not given
        if allowedvalues is not None:
            raise ValueError('allowedvalues not allowed for arrays')

        super().__init__(name, record, defaultvalue=defaultvalue,
                         valuerequired=valuerequired, allowedvalues=allowedvalues,
                         metadatakey=metadatakey, metadataparent=metadataparent,
                         modelpath=modelpath, description=description)

    def set_value_mod(self, val):
        if val is None:
            return None
        
        elif isinstance(val, str):
            val = val.strip().split()

        return np.asarray(val, dtype=int)
    
    def build_model_value(self):
        return self.value.tolist()
        
    def load_model_value(self, val):
        return np.asarray(val, dtype=int)
        


from typing import Optional, Union, Any

from DataModelDict import DataModelDict as DM

import numpy as np

from ..query import load_query
from . import Value

class IntValue(Value):
    
    def set_value_mod(self, val):

        # Check if value is in #text
        val = self.set_value_mod_textfield(val)

        if val is None:
            return None
        elif isinstance(val, float):
            rval = round(val)
            if not np.isclose(rval, val):
                raise TypeError(f'{self.name} not an int!')
            return rval
        else:
            try:
                return int(val)
            except Exception as e:
                raise ValueError(f'{self.name} not an int!') from e
    
    @property
    def _default_queries(self) -> dict:
        """dict: Default query operations to associate with the Parameter style"""
        if self.metadatakey is False:
            return {}
        else:
            return {
                self.name: load_query('int_match',
                                    name=self.metadatakey,
                                    parent=self.metadataparent,
                                    path=f'{self.record.modelroot}.{self.modelpath}',
                                    description=f'Return only the records where {self.description} matches a given value')
            }


from typing import Optional, Union, Any
import datetime

from DataModelDict import DataModelDict as DM

import numpy as np

from ..query import load_query
from . import Value

class DateValue(Value):
    
    def set_value_mod(self, val):
        
        # Check if value is in #text
        val = self.set_value_mod_textfield(val)
        
        if val is None:
            val = datetime.date.today()
        elif not isinstance(val, datetime.date):
            val = datetime.datetime.strptime(val, '%Y-%m-%d').date()
        return val
        
    def build_model_value(self):
        return str(self.value)

    @property
    def _default_queries(self) -> dict:
        """dict: Default query operations to associate with the Parameter style"""
        if self.metadatakey is False:
            return {}
        else:
            return {
                self.name: load_query('date_match',
                                    name=self.metadatakey,
                                    parent=self.metadataparent,
                                    path=f'{self.record.modelroot}.{self.modelpath}',
                                    description=f'Return only the records where {self.description} matches a given value')
            }
    


@property
def g(self) -> float:
    """gram"""
    return 1e-3 * self.kg

@property
def mg(self) -> float:
    """milligram"""
    return 1e-3 * self.g

@property
def ug(self) -> float:
    """microgram"""
    return 1e-6 * self.g

@property
def ng(self) -> float:
    """nanogram"""
    return 1e-9 * self.g

@property
def pg(self) -> float:
    """picogram"""
    return 1e-12 * self.g

@property
def fg(self) -> float:
    """femtogram"""
    return 1e-15 * self.g

@property
def tonne(self) -> float:
    """tonne"""
    return 1e3 * self.kg

@property
def amu(self) -> float:
    """atomic mass unit"""
    return 1.66053906660e-27 * self.kg

@property
def Da(self) -> float:
    """dalton"""
    return self.amu

@property
def kDa(self) -> float:
    """kilodalton"""
    return 1e3 * self.Da

@property
def lbm(self) -> float:
    """pound mass (international avoirdupois pound)"""
    return 0.45359237 * self.kg


# coding: utf-8

# iprPy imports
from . import recordmanager, databasemanager, querymanager, valuemanager

__all__ = ['check_modules']

def check_modules():
    """
    Prints lists of the modular components that were successfully and
    unsuccessfully loaded.
    """
    databasemanager.check_styles()
    print()
    recordmanager.check_styles()
    print()
    querymanager.check_styles()
    print()
    valuemanager.check_styles()
    print()




@property
def Gal(self) -> float:
    """galileo"""
    return 1. * self.cm / self.s**2

@property
def mGal(self) -> float:
    """milligalileo"""
    return 1e-3 * self.Gal

@property
def uGal(self) -> float:
    """microgalileo"""
    return 1e-6 * self.Gal

@property
def eotvos(self) -> float:
    """eotvos"""
    return 1e-9 / self.s**2


@property
def J(self) -> float:
    """joule"""
    return (self.kg * self.m**2) / self.s**2

@property
def mJ(self) -> float:
    """milligram"""
    return 1e-3 * self.J

@property
def uJ(self) -> float:
    """microgram"""
    return 1e-6 * self.J

@property
def nJ(self) -> float:
    """nanogram"""
    return 1e-9 * self.J

@property
def pJ(self) -> float:
    """picogram"""
    return 1e-12 * self.J

@property
def fJ(self) -> float:
    """femtogram"""
    return 1e-15 * self.J

@property
def kJ(self) -> float:
    """kilojoule"""
    return 1e3 * self.J

@property
def MJ(self) -> float:
    """megajoule"""
    return 1e6 * self.J

@property
def GJ(self) -> float:
    """gigajoule"""
    return 1e9 * self.J

@property
def erg(self) -> float:
    """erg"""
    return 1e-7 * self.J

@property
def eV(self) -> float:
    """electron volt"""
    return 1.602176634e-19 * self.J

@property
def meV(self) -> float:
    """millielectron volt"""
    return 1e-3 * self.eV

@property
def keV(self) -> float:
    """kiloelectron volt"""
    return 1e3 * self.eV

@property
def MeV(self) -> float:
    """megaelectron volt"""
    return 1e6 * self.eV

@property
def GeV(self) -> float:
    """gigaelectron volt"""
    return 1e9 * self.eV

@property
def TeV(self) -> float:
    """teraelectron volt"""
    return 1e12 * self.eV

@property
def btu(self) -> float:
    """British thermal unit"""
    return 1055.06 * self.J

@property
def smallcal(self) -> float:
    """small calorie (a.k.a. gram calorie)"""
    return 4.184 * self.J

@property
def kcal(self) -> float:
    """kilocalorie (a.k.a large Calorie, dietary Calorie)"""
    return 4184. * self.J

@property
def Wh(self) -> float:
    """watt-hour"""
    return 3600. * self.J

@property
def kWh(self) -> float:
    """kilowatt-hour"""
    return 1e3 * self.Wh



@property 
def ohm(self) -> float:
    """ohm"""
    return self.V / self.A

@property
def Ω(self) -> float:
    """ohm"""
    return self.ohm

@property
def mohm(self) -> float:
    """milliohm"""
    return 1e-3 * self.ohm

@property
def mΩ(self) -> float:
    """milliohm"""
    return self.mohm

@property
def kohm(self) -> float:
    """kiloohm"""
    return 1e3 * self.ohm

@property
def kΩ(self) -> float:
    """kiloohm"""
    return self.kohm

@property
def Mohm(self) -> float:
    """megaohm"""
    return 1e6 * self.ohm

@property
def MΩ(self) -> float:
    """megaohm"""
    return self.Mohm

@property
def Gohm(self) -> float:
    """gigaohm"""
    return 1e9 * self.ohm

@property
def GΩ(self) -> float:
    """gigaohm"""
    return self.Gohm

@property
def S(self) -> float:
    """siemens"""
    return 1. / self.ohm

@property
def mS(self) -> float:
    """millisiemens"""
    return 1e-3 * self.S

@property
def uS(self) -> float:
    """microsiemens"""
    return 1e-6 * self.S

@property
def nS(self) -> float:
    """nanosiemens"""
    return 1e-9 * self.S


# coding: utf-8
# Standard Python libraries
from pathlib import Path
import ast
import shutil
import tarfile
from typing import Optional, Tuple, Union

# http://www.numpy.org/
import numpy as np

# https://pandas.pydata.org/
import pandas as pd

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

# iprPy imports
from ..tools import aslist, iaslist
from . import Database
from ..record import recordmanager, load_record, Record

class LocalDatabase(Database):

    def __init__(self,
                 host: str,
                 format: str = 'json',
                 indent: Optional[int] = None):
        """
        Initializes a connection to a local database of JSON/XML records
        stored in a local directory.
        
        Parameters
        ----------
        host : str
            The host name (local directory path) for the database.
        format : str, optional
            The format that the model records are saved as.  Can be either
            JSON or XML.  Default value is JSON.
        indent : int or None, optional
            The indentation used when saving the records.  If None (default)
            then the saved records are compact.  Otherwise, the lines in the
            file will be indented by multiples of this value based on the
            model's element recursion.
        """
        # Make the path if needed
        host = Path(host)
        if not host.is_dir():
            host.mkdir(parents=True)

        # Get absolute path to host
        host = host.resolve()

        # Pass host to Database initializer
        Database.__init__(self, host)

        # Set default format and indent values
        self.__format = format
        self.__indent = indent

    @property
    def style(self) -> str:
        """str: The database style"""
        return 'local'

    @property
    def format(self) -> str:
        """str: The format that records are saved as: 'json' or 'xml'"""
        return self.__format

    @property
    def indent(self) -> Optional[int]:
        """int or None: The record indentation setting to use when saving records."""
        return self.__indent

    def cache(self,
              style: str,
              refresh: bool = False,
              addnew: bool = True) -> pd.DataFrame:
        """
        Loads/generates the metadata cache csv file for a given record style.

        Parameters
        ----------
        style : str 
            The record style to retrieve the metadata content for.
        refresh : bool, optional
            If True, then the metadata content will be rebuilt by loading every
            record of the given style.  If False (default), the stored metadata
            for records will be used rather than loading from the files.
        addnew : bool, optional
            If True (default), then the metadata for new records will be
            appended to the stored metadata cache.  If False, then the stored
            metadata is returned as is.

        Returns
        -------
        pandas.DataFrame
            The contents of the cache csv file.
        """
        recordmanager.assert_style(style)
        cachefile = Path(self.host, f'{style}.csv')

        if cachefile.is_file() and refresh is False:

            # Load cache file
            cache = pd.read_csv(cachefile)

            def interpret(series, key):
                """Safely convert dict and list elements from str"""
                try:
                    assert series[key][0] in '{['
                    return ast.literal_eval(series[key])
                except:
                    return series[key]

            def toint(column):
                """Convert int columns as needed"""
                try:
                    assert column.dtype == float
                    newcolumn = column.astype(int)
                    assert np.allclose(column, newcolumn, atol=0.0)
                except:
                    return column
                else:
                    return newcolumn

            # Interpret int, dict and list elements
            if len(cache) > 0:
                cache = cache.apply(toint, axis=0)
                for key in cache.keys():
                    cache[key] = cache.apply(interpret, axis=1, args=[key])
            else:
                r = load_record(style)
                cache = pd.DataFrame(columns=r.metadatakeys)

        else:
            # Initialize new cache
            r = load_record(style)
            cache = pd.DataFrame(columns=r.metadatakeys)

        if addnew is True:

            # Compare names in the cache to file names in the directory
            cachenames = set(cache.name)
            filenames = set([fname.stem for fname in Path(self.host, style).glob(f'*.{self.format}')])
            newnames = filenames.difference(cachenames)
            deletednames = cachenames.difference(filenames)

            # Load new entries
            if len(newnames) > 0:
                newrecords = []
                for name in newnames:
                    fname = Path(self.host, style, f'{name}.{self.format}')
                    record = load_record(style, model=fname, name=name)
                    newrecords.append(record.metadata())
                newrecords = pd.DataFrame(newrecords)
                if not cache.empty:
                    cache = pd.concat([cache, newrecords], sort=False).sort_values('name')
                else:
                    cache = newrecords.sort_values('name')
                refresh = True

            # Delete missing entries
            if len(deletednames) > 0:
                cache = cache[~cache.name.isin(deletednames)]
                refresh = True

        # Refresh cache file
        if refresh:
            cache.to_csv(cachefile, index=False)

        return cache

    def get_records(self, 
                    style: Optional[str] = None,
                    return_df: bool = False,
                    refresh_cache: bool = False,
                    **kwargs) -> Union[list, Tuple[list, pd.DataFrame]]:
        """
        Produces a list of all matching records in the database.
        
        Parameters
        ----------
        style : str, optional
            The record style to search.
        refresh_cache : bool, optional
            Indicates if the metadata cache file is to be refreshed.  If False,
            metadata for new records will be added but the old record metadata
            fields will not be updated.  If True, then the metadata for all
            records will be regenerated, which is needed to update the metadata
            for modified records.
        return_df : bool, optional
            If True, then the corresponding pandas.Dataframe of metadata
            will also be returned
        **kwargs : any, optional
            Any of the record-specific metadata keywords that can be searched
            for.

        Returns
        ------
        records : numpy.NDArray
            All records from the database matching the given parameters.
        records_df : pandas.DataFrame
            The corresponding metadata values for the records.  Only returned
            if return_df is True.
        """
        # Set default search parameters
        if style is None:
            style = self.select_record_style()

        # Get df
        df = self.get_records_df(style, refresh_cache=refresh_cache, **kwargs)

        # Load only the matching records
        records = []
        if len(df) > 0:
            for name in df.name:
                fname = Path(self.host, style, f'{name}.{self.format}')
                records.append(load_record(style, model=fname, database=self))

        records = np.array(records)

        if return_df:
            return records, df
        else:
            return records

    def get_records_df(self, 
                       style: Optional[str] = None,
                       refresh_cache: bool = False,
                       **kwargs) -> pd.DataFrame:
        """
        Produces a table of metadata for matching records in the database.
        
        Parameters
        ----------
        style : str, optional
            The record style to limit the search by.
        refresh_cache : bool, optional
            Indicates if the metadata cache file is to be refreshed.  If False,
            metadata for new records will be added but the old record metadata
            fields will not be updated.  If True, then the metadata for all
            records will be regenerated, which is needed to update the metadata
            for modified records.
        **kwargs : any, optional
            Any of the record-specific metadata keywords that can be searched
            for.
        
        Returns
        ------
        records_df : pandas.DataFrame
            The corresponding metadata values for the records.
        """

        # Set default search parameters
        if style is None:
            style = self.select_record_style()

        if 'name' in kwargs and kwargs['name'] is not None:
            # Load named records
            cache = []
            for name in aslist(kwargs['name']):
                fname = Path(self.host, style, f'{name}.{self.format}')
                if fname.exists():
                    record = load_record(style, model=fname)
                    cache.append(record.metadata())

            # Build cache DataFrame
            if len(cache) == 0:
                # Initialize new cache
                r = load_record(style)
                cache = pd.DataFrame(columns=r.metadatakeys)
            else:
                cache = pd.DataFrame(cache)

        else:
            # Load cache file
            cache = self.cache(style, refresh=refresh_cache)

        # Filter based on the record's pandasfilter method
        mask = load_record(style).pandasfilter(cache, **kwargs)
        df = cache[mask].reset_index(drop=True)

        return df

    def get_record(self,
                   style: Optional[str] = None,
                   refresh_cache: bool = False,
                   **kwargs) -> Record:
        """
        Returns a single matching record from the database.
        
        Parameters
        ----------
        style : str, optional
            The record style to limit the search by.
        refresh_cache : bool, optional
            Indicates if the metadata cache file is to be refreshed.  If False,
            metadata for new records will be added but the old record metadata
            fields will not be updated.  If True, then the metadata for all
            records will be regenerated, which is needed to update the metadata
            for modified records.
        **kwargs : any, optional
            Any of the record-specific metadata keywords that can be searched
            for.
        
        Returns
        -------
        Record
            The single record from the database matching the given parameters.
        
        Raises
        ------
        ValueError
            If multiple or no matching records found.
        """

        if style is None:
            styles = recordmanager.loaded_style_names
        else:
            styles = aslist(style)

        # Get records
        records = []
        for style in styles:
            records.append(self.get_records(style, refresh_cache=refresh_cache, **kwargs))
        records = np.hstack(records)

        # Verify that there is only one matching record
        if len(records) == 1:
            return records[0]
        elif len(records) == 0:
            raise ValueError('No matching records found')
        else:
            raise ValueError('Multiple matching records found')

    def count_records(self,
                      style: Optional[str] = None,
                      **kwargs) -> int:
        """
        Retrieves a count of matching records from the database.  Fast if you
        only want the total number of records of a given style. 
        
        Parameters
        ----------
        style : str, optional
            The record style to search.
        **kwargs : any, optional
            Any of the record-specific metadata keywords that can be searched
            for, or other kwargs supported by get_records_df.  Note that if
            you give kwargs the method will be much slower as it is effectively
            calling len(get_records_df).
            
        Returns
        ------
        int
            The count of records in the database matching the given parameters.
        """
        # Set default search parameters
        if style is None:
            style = self.select_record_style()

        # Fast count of all records of a given style
        if len(kwargs) == 0:
            count = len([i for i in Path(self.host, style).glob(f'*.{self.format}')])

        # Use get_records_df for delimited searches    
        else:
            count = len(self.get_records_df(style, **kwargs))

        return count

    def add_record(self,
                   record: Optional[Record] = None,
                   style: Optional[str] = None,
                   name: Optional[str] = None,
                   model: Union[str, DM, None] = None,
                   build: bool = False,
                   verbose: bool = False) -> Record:
        """
        Adds a new record to the database.
        
        Parameters
        ----------
        record : Record, optional
            The new record to add to the database.  If not given, then name,
            style and model are required.
        style : str, optional
            The record style for the new record.  Required if record is not
            given.
        name : str, optional
            The name to assign to the new record.  Required if record is not
            given.
        model : str or DataModelDict, optional
            The model contents of the new record.  Required if record is not
            given.
        build : bool, optional
            If True, then the uploaded content will be (re)built based on the
            record's attributes.  If False (default), then record's existing
            content will be loaded if it exists, or built if it doesn't exist.
        verbose : bool, optional
            If True, info messages will be printed during operations.  Default
            value is False.

        Returns
        ------
        Record
            Either the given record or a record composed of the name, style,
            and model.
        
        Raises
        ------
        ValueError
            If style, name and/or model given with record, or a matching record
            already exists.
        """

        # Create Record object if not given
        if record is None:
            record = load_record(style, model=model, name=name)

        # Issue a ValueError for competing kwargs
        elif style is not None or name is not None or model is not None:
            raise ValueError('kwargs style, name, and model cannot be given with kwarg record')

        # Verify that there isn't already a record with a matching name
        style_dir = Path(self.host, record.style)
        fname = Path(style_dir, f'{record.name}.{self.format}')
        if fname.is_file():
            raise ValueError(f'Record {record.name} already exists')

        # Make record style directory if needed
        if not style_dir.is_dir():
            style_dir.mkdir()

        # Retrieve/build model contents
        try:
            assert build is False
            model = record.model
        except:
            model = record.build_model()

        # Save record
        with open(fname, 'w', encoding='UTF-8') as f:
            if self.format == 'json':
                model.json(fp=f, indent=self.indent, ensure_ascii=False)
            elif self.format == 'xml':
                model.xml(fp=f, indent=self.indent)

        if verbose:
            print(f'{record} added to {self.host}')

        return record

    def update_record(self,
                      record: Optional[Record] = None,
                      style: Optional[str] = None,
                      name: Optional[str] = None,
                      model: Union[str, DM, None] = None,
                      build: bool = False,
                      verbose: bool = False) -> Record:
        """
        Replaces an existing record with a new record of matching name and
        style, but new content.
        
        Parameters
        ----------
        record : Record, optional
            The record with new content to update in the database.  If not
            given, content is required along with name and/or style to
            uniquely define a record to update.
        style : str, optional
            The style of the record to update.
        name : str, optional
            The name to uniquely identify the record to update.
        model : str or DataModelDict, optional
            The model contents of the new record.  Required if record is not
            given.
        build : bool, optional
            If True, then the uploaded content will be (re)built based on the
            record's attributes.  If False (default), then record's existing
            content will be loaded if it exists, or built if it doesn't exist.
        verbose : bool, optional
            If True, info messages will be printed during operations.  Default
            value is False.

        Returns
        ------
        Record
            Either the given record or a record composed of the name, style,
            and content.
        
        Raises
        ------
        TypeError
            If no new content is given.
        ValueError
            If style and/or name content given with record.
        """

        # Create Record object if not given
        if record is None:
            if model is None:
                raise TypeError('no new model given')

            record = load_record(style, model=model, name=name)

        # Issue a ValueError for competing kwargs
        elif style is not None or name is not None:
            raise ValueError('kwargs style and name cannot be given with kwarg record')

        # Replace content in record object
        elif model is not None:
            record = load_record(record.style, model=model, name=record.name)

        # Check if record already exists
        style_dir = Path(self.host, record.style)
        fname = Path(style_dir, f'{record.name}.{self.format}')
        if not fname.is_file():
            raise ValueError(f'No existing record {record.name} found')

        # Retrieve/build model contents
        try:
            assert build is False
            model = record.model
        except:
            model = record.build_model()

        # Save record
        with open(fname, 'w', encoding='UTF-8') as f:
            if self.format == 'json':
                model.json(fp=f, indent=self.indent, ensure_ascii=False)
            elif self.format == 'xml':
                model.xml(fp=f, indent=self.indent)

        if verbose:
            print(f'{record} updated in {self.host}')

        return record

    def delete_record(self,
                      record: Optional[Record] = None,
                      style: Optional[str] = None,
                      name: Optional[str] = None,
                      verbose: bool = False):
        """
        Permanently deletes a record from the database.
        
        Parameters
        ----------
        record : Record, optional
            The record to delete from the database.  If not given, name and/or
            style are needed to uniquely define the record to delete.
        name : str, optional
            The name of the record to delete.
        style : str, optional
            The style of the record to delete.
        verbose : bool, optional
            If True, info messages will be printed during operations.  Default
            value is False.
            
        Raises
        ------
        ValueError
            If style and/or name given with record.
        """

        # Create Record object if not given
        if record is None:
            record = self.get_record(name=name, style=style)

        # Issue a ValueError for competing kwargs
        elif style is not None or name is not None:
            raise ValueError('kwargs style and name cannot be given with kwarg record')

         # Delete record file
        fname = Path(self.host, record.style, f'{record.name}.{self.format}')
        if fname.is_file():
            fname.unlink()
        else:
            raise ValueError(f'No existing {record.style} record {record.name} found')

        if verbose:
            print(f'{record} deleted from {self.host}')

    def add_tar(self, 
                record: Optional[Record] = None,
                style: Optional[str] = None,
                name: Optional[str] = None,
                tar: Optional[bytes] = None,
                root_dir: Optional[Path] = None):
        """
        Archives and stores a folder associated with a record.
        
        Parameters
        ----------
        record : Record, optional
            The record to associate the tar archive with.  If not given, then
            name and/or style necessary to uniquely identify the record are
            needed.
        name : str, optional
            The name to use in uniquely identifying the record.
        style : str, optional
            The style to use in uniquely identifying the record.
        tar : bytes, optional
            The bytes content of a tar file to save.  tar cannot be given
            with root_dir.
        root_dir : str, optional
            Specifies the root directory for finding the directory to archive.
            The directory to archive is at <root_dir>/<name>.  (Default is to
            set root_dir to the current working directory.)  tar cannot be given
            with root_dir.
        
        Raises
        ------
        ValueError
            If style and/or name content given with record or the record already
            has a folder or a tar archive.
        """

        # Create Record object if not given
        if record is None:
            record = self.get_record(name=name, style=style)

        # Issue a ValueError for competing kwargs
        elif style is not None or name is not None:
            raise ValueError('kwargs style and name cannot be given with kwarg record')

        # Verify that record exists
        else:
            record = self.get_record(name=record.name, style=record.style)

        # Build path to record
        dir_path = Path(self.host, record.style, record.name)
        tar_path = Path(self.host, record.style, f'{record.name}.tar.gz')

        # Check if an archive or folder already exists
        if tar_path.exists():
            raise ValueError('Record already has an archive')
        elif dir_path.exists():
            raise ValueError('Record already has a folder')

        # Make archive
        if tar is None:
            if root_dir is None:
                root_dir = '.'
            target = Path(root_dir, record.name)

            tar = tarfile.open(tar_path, 'w:gz')
            tar.add(target, target.name)
            tar.close()

        elif root_dir is None:
            with open(tar_path, 'wb') as f:
                f.write(tar)
        else:
            raise ValueError('tar and root_dir cannot both be given')

    def get_tar(self,
                record: Optional[Record] = None,
                style: Optional[str] = None,
                name: Optional[str] = None,
                raw: bool = False) -> Union[tarfile.TarFile, bytes]:
        """
        Retrieves the tar archive associated with a record in the database.
        
        Parameters
        ----------
        record : Record, optional
            The record to retrieve the associated tar archive for.
        name : str, optional
            The name to use in uniquely identifying the record.
        style : str, optional
            The style to use in uniquely identifying the record.
        raw : bool, optional
            If True, return the archive as raw binary content. If 
            False, return as an open tarfile. (Default is False)
        
        Returns
        -------
        tarfile or bytes
            The tar archive as an open tarfile if raw=False, or as a binary str if
            raw=True.
        
        Raises
        ------
        ValueError
            If style and/or name content given with record.
        """

        # Create Record object if not given
        if record is None:
            record = self.get_record(name=name, style=style)

        # Issue a ValueError for competing kwargs
        elif style is not None or name is not None:
            raise ValueError('kwargs style and name cannot be given with kwarg record')

        # Build path to record
        tar_path = Path(self.host, record.style, record.name+'.tar.gz')

        # Return content
        if raw is True:
            with open(tar_path, 'rb') as f:
                return f.read()
        else:
            tar = tarfile.open(tar_path)
            record.tar = tar
            return tar

    def delete_tar(self, 
                   record: Optional[Record] = None,
                   style: Optional[str] = None,
                   name: Optional[str] = None):
        """
        Deletes a tar file from the database.
        
        Parameters
        ----------
        record : Record, optional
            The record associated with the tar archive to delete.  If not
            given, then name and/or style necessary to uniquely identify
            the record are needed.
        name : str, optional
            The name to use in uniquely identifying the record.
        style : str, optional
            The style to use in uniquely identifying the record.
        
        Raises
        ------
        ValueError
            If style and/or name content given with record.
        """

        # Create Record object if not given
        if record is None:
            record = self.get_record(name=name, style=style)

        # Issue a ValueError for competing kwargs
        elif style is not None or name is not None:
            raise ValueError('kwargs style and name cannot be given with kwarg record')

        # Build path to tar file
        tar_path = Path(self.host, record.style, record.name+'.tar.gz')

        # Delete record if it exists
        if tar_path.is_file():
            tar_path.unlink()

    def update_tar(self,
                   record: Optional[Record] = None,
                   style: Optional[str] = None,
                   name: Optional[str] = None,
                   tar: Optional[bytes] = None,
                   root_dir: Optional[Path] = None):
        """
        Replaces an existing tar archive for a record with a new one.
        
        Parameters
        ----------
        record : Record, optional
            The record to associate the tar archive with.  If not given, then 
            name and/or style necessary to uniquely identify the record are 
            needed.
        name : str, optional
            The name to use in uniquely identifying the record.
        style : str, optional
            The style to use in uniquely identifying the record.
        tar : bytes, optional
            The bytes content of a tar file to save.  tar cannot be given
            with root_dir.
        root_dir : str, optional
            Specifies the root directory for finding the directory to archive.
            The directory to archive is at <root_dir>/<name>.  (Default is to
            set root_dir to the current working directory.)  tar cannot be given
            with root_dir.
        """

        # Delete the existing tar archive stored in the database
        self.delete_tar(record=record, name=name)

        # Add the new tar archive
        self.add_tar(record=record, name=name, style=style, tar=tar,
                     root_dir=root_dir)

    def add_folder(self,
                   record: Optional[Record] = None,
                   style: Optional[str] = None,
                   name: Optional[str] = None,
                   filenames: Union[str, Path, list, None] = None,
                   root_dir: Optional[Path] = None):
        """
        Stores a folder associated with a record.
        
        Parameters
        ----------
        record : Record, optional
            The record to associate the folder with.  If not given, then
            name and/or style necessary to uniquely identify the record are
            needed.
        style : str, optional
            The style to use in uniquely identifying the record.
        name : str, optional
            The name to use in uniquely identifying the record.
        filenames : str or list, optional
            The paths to files that are to be added to the record's folder.
            Cannot be given with root_dir.
        root_dir : str, optional
            Specifies the root directory for finding the folder to add.
            The directory to add is at <root_dir>/<name>.  (Default is to
            set root_dir to the current working directory.)  Cannot be given
            with filenames.
        
        Raises
        ------
        ValueError
            If style and/or name content given with record or the record already
            has a folder or a tar archive.
        """

        # Create Record object if not given
        if record is None:
            record = self.get_record(name=name, style=style)

        # Issue a ValueError for competing kwargs
        elif style is not None or name is not None:
            raise ValueError('kwargs style and name cannot be given with kwarg record')

        # Verify that record exists
        else:
            record = self.get_record(name=record.name, style=record.style)

        # Build database paths
        dir_path = Path(self.host, record.style, record.name)
        tar_path = Path(self.host, record.style, f'{record.name}.tar.gz')

        # Check if an archive or folder already exists
        if tar_path.exists():
            raise ValueError('Record already has an archive')
        elif dir_path.exists():
            raise ValueError('Record already has a folder')

        # Copy folder
        if filenames is None:
            if root_dir is None:
                root_dir = '.'
            source = Path(root_dir, record.name)
            shutil.copytree(source, dir_path)

        # Copy files
        elif root_dir is None:
            dir_path.mkdir(parents=True)
            for filename in iaslist(filenames):
                shutil.copy2(filename, Path(dir_path, Path(filename).name))

        else:
            raise ValueError('filenames and root_dir cannot both be given')

    def get_folder(self,
                   record: Optional[Record] = None,
                   style: Optional[str] = None,
                   name: Optional[str] = None) -> Path:
        """
        Retrives the location of the folder associated with a record in the
        database. 
        
        Parameters
        ----------
        record : Record, optional
            The record to retrieve the associated folder location for.
        style : str, optional
            The style to use in uniquely identifying the record.
        name : str, optional
            The name to use in uniquely identifying the record.
        
        Returns
        -------
        pathlib.Path
            The path to the record's folder
        
        Raises
        ------
        ValueError
            If style and/or name content given with record.
        NotADirectoryError
            If the record's folder doesn't exist.
        """

        # Create Record object if not given
        if record is None:
            record = self.get_record(name=name, style=style)

        # Issue a ValueError for competing kwargs
        elif style is not None or name is not None:
            raise ValueError('kwargs style and name cannot be given with kwarg record')

        # Verify that record exists
        else:
            record = self.get_record(name=record.name, style=record.style)

        # Build path to folder
        dir_path = Path(self.host, record.style, record.name)

        # Return path
        if dir_path.exists():
            return dir_path
        else:
            raise NotADirectoryError('No folder saved for the record')

    def delete_folder(self,
                      record: Optional[Record] = None,
                      style: Optional[str] = None,
                      name: Optional[str] = None):
        """
        Deletes a folder from the database.
        
        Parameters
        ----------
        record : Record, optional
            The record associated with the folder to delete.  If not
            given, then name and/or style necessary to uniquely identify
            the record are needed.
        style : str, optional
            The style to use in uniquely identifying the record.
        name : str, optional
            The name to use in uniquely identifying the record.
        
        Raises
        ------
        ValueError
            If style and/or name content given with record.
        """

        # Create Record object if not given
        if record is None:
            record = self.get_record(name=name, style=style)

        # Issue a ValueError for competing kwargs
        elif style is not None or name is not None:
            raise ValueError('kwargs style and name cannot be given with kwarg record')

        # Verify that record exists
        else:
            record = self.get_record(name=record.name, style=record.style)

        # Build path to tar file
        dir_path = Path(self.host, record.style, record.name)

        # Delete record if it exists
        if dir_path.exists():
            shutil.rmtree(dir_path)

    def update_folder(self,
                      record: Optional[Record] = None,
                      style: Optional[str] = None,
                      name: Optional[str] = None,
                      filenames: Union[str, Path, list, None] = None,
                      root_dir: Optional[Path] = None,
                      clear: bool = True):
        """
        Updates an existing folder for a record.
        
        Parameters
        ----------
        record : Record, optional
            The record to associate the folder with.  If not given, then 
            name and/or style necessary to uniquely identify the record are 
            needed.
        style : str, optional
            The style to use in uniquely identifying the record.
        name : str, optional
            The name to use in uniquely identifying the record.
        filenames : str or list, optional
            The paths to files that are to be added to the record's folder.
            Cannot be given with root_dir.
        root_dir : str, optional
            Specifies the root directory for finding the folder to add.
            The directory to add is at <root_dir>/<name>.  (Default is to
            set root_dir to the current working directory.)  Cannot be given
            with filenames.
        clear : bool, optional
            If True (default), then the current folder contents will be deleted
            before the new contents are added.  If False, existing files may
            remain if the new content does not overwrite it.
        """
        # Check competing parameters
        if filenames is not None and root_dir is not None:
            raise ValueError('filenames and root_dir cannot both be given')

        # Create Record object if not given
        if record is None:
            record = self.get_record(name=name, style=style)

        # Issue a ValueError for competing kwargs
        elif style is not None or name is not None:
            raise ValueError('kwargs style and name cannot be given with kwarg record')

        # Verify that record exists
        else:
            record = self.get_record(name=record.name, style=record.style)

        # Build database paths
        dir_path = Path(self.host, record.style, record.name)

        # Delete existing folder
        if clear is True:
            if dir_path.exists():
                shutil.rmtree(dir_path)

        # Copy folder
        if filenames is None:
            if root_dir is None:
                root_dir = '.'
            source = Path(root_dir, record.name)
            shutil.copytree(source, dir_path)

        # Copy files
        elif root_dir is None:
            dir_path.mkdir(parents=True)
            for filename in iaslist(filenames):
                shutil.copy2(filename, Path(dir_path, Path(filename).name))



# coding: utf-8

# Standard Python libraries
from typing import Any, Optional

import pandas as pd

# Relative imports
from ..tools import aslist

from .Query import Query

class DummyQuery(Query):
    """Class for a query field that doesn't perform a database query"""

    def __init__(self,
                 name: Optional[str] = None,
                 parent: Optional[str] = None,
                 path: Optional[str] = None,
                 description: str = '',
                 value: Optional[Any] = None):
        """
        Query initialization

        Parameters
        ----------
        name : str or None, optional
            The metadata key associated with the data field.  Must be set
            to use the pandas query method.
        parent : str or None, optional
            Allows for the pandas query operations to work on embedded
            metadata dicts.  If given, the pandas query method will check the
            value of metadata[parent][name].
        path : str or None, optional
            The record data path to the data field.  Levels are delimited by
            periods.  Must be given to use the mongo query method.
        description : str, optional
            Description of the query operation, i.e. what it is searching.
        value : any, optional
            If None (default), then the query operations will do nothing
            (i.e. always return True values). If a different value is
            given, then the value(s) provided during the query call will
            be checked against this value for a match to determine if all
            True or all False are returned by the query operation.
        """
        super().__init__(name, parent, path, description)
        self.__value = value

    @property
    def style(self) -> str:
        """str: The query style"""
        return 'dummy'

    @property
    def parameter_type(self) -> str:
        """str: The types of query parameter values accepted by this query style"""
        return 'any, optional'

    @property
    def value(self) -> Any:
        """Any: A single constant value that will always"""
        return self.__value
    
    def mongo(self,
              querylist: list,
              value: Any,
              prefix: str = ''):
        """
        Builds a Mongo query operation for the field.

        Parameters
        ----------
        querylist : list
            The working list of mongo query operations which is to be appended
            with the operation for this query object.
        value : any
            The value of the field to query on.  If None, then no new query
            operation will be added.
        prefix : str, optional
            An optional prefix to add before the query path.  Used by Record's
            mongoquery to start each path with "content."
        """
        # Build an always false query if self.value not in value
        if value is not None and self.value is not None:
            if self.value not in aslist(value):
                querylist.append( {'nonexistantpath': {'$exists': True} } )

    def pandas(self,
               df: pd.DataFrame,
               value: Any) -> pd.Series:
        """
        Applies a query filter to the metadata for the field.
        
        Parameters
        ----------
        df : pandas.DataFrame
            A table of metadata for multiple records of the record style.
        value : any
            The value of the field to query on.  If None, then it should return
            True for all rows of df.
        
        Returns
        -------
        pandas.Series
            Boolean map of matching values
        """
        # Construct all true values
        alltrue = pd.Series(True, df.index, dtype=bool)

        # Return all false if self.value not in value
        if value is not None and self.value is not None:        
            if self.value not in aslist(value):
                return ~alltrue

        # Otherwise return all true
        return alltrue



from typing import Optional, Union, Any

from DataModelDict import DataModelDict as DM

from ..query import load_query
from . import Value

class LongStrValue(Value):
    
    def set_value_mod(self, val):

        # Check if value is in #text
        val = self.set_value_mod_textfield(val)

        # Pass None values through
        if val is None:
            return None
            
        return str(val)
    
    @property
    def _default_queries(self) -> dict:
        """dict: Default query operations to associate with the Parameter style"""
        if self.metadatakey is False:
            return {}
        else:
            return {
                self.name: load_query('str_contains',
                                    name=self.metadatakey,
                                    parent=self.metadataparent,
                                    path=f'{self.record.modelroot}.{self.modelpath}',
                                    description=f'Return only the records where {self.description} contains the given values')
            }


from typing import Optional, Union, Any

from DataModelDict import DataModelDict as DM

from ..query import load_query
from . import Value
from ..tools import iaslist

class StrListValue(Value):
    
    def set_value_mod(self, val):
        if val is None:
            return None
        else:
            return [str(v) for v in iaslist(val)]
    
    @property
    def _default_queries(self) -> dict:
        """dict: Default query operations to associate with the Parameter style"""
        if self.metadatakey is False:
            return {}
        else:
            return {
                self.name: load_query('list_contains',
                                    name=self.metadatakey,
                                    parent=self.metadataparent,
                                    path=f'{self.record.modelroot}.{self.modelpath}',
                                    description=f'Return only the records where {self.description} contains a given value')
            }
    
    def build_model_value(self):
        if self.value is None or len(self.value) == 0:
            return
        elif len(self.value) == 1:
            return self.value[0]
        return self.value


# coding: utf-8
# Standard Python libraries
from importlib import resources

# Read version from VERSION file
if hasattr(resources, 'files'):
    __version__ = resources.files('yabadaba').joinpath('VERSION').read_text(encoding='UTF-8')
else:
    __version__ = resources.read_text('yabadaba', 'VERSION', encoding='UTF-8').strip()

# Relative imports
from .UnitConverter import unitconvert
from . import tools
from .Settings import settings

from . import query
from .query import querymanager, load_query

from . import value
from .value import valuemanager, load_value

from . import record
from .record import recordmanager, load_record

from . import database
from .database import databasemanager, load_database

from .check_modules import check_modules
from .querydoc import querydoc

__all__ = ['__version__', 'tools', 'settings', 'unitconvert',
           'query', 'load_query', 'querymanager',
           'record', 'load_record', 'recordmanager',
           'value', 'load_value', 'valuemanager',
           'database', 'load_database', 'databasemanager',
           'check_modules', 'querydoc']
__all__.sort()



from typing import Optional, Union, Any

from DataModelDict import DataModelDict as DM

from . import recordmanager, Record
from ..query import load_query
from ..value import Value
from ..tools import aslist, iaslist

class RecordValue(Value):
    
    def __init__(self,
                 name: str,
                 record,
                 recordclass: Union[str, type[Record]],
                 defaultvalue: Optional[Any] = None,
                 valuerequired: bool = False,
                 allowedvalues: Optional[tuple] = None,
                 metadatakey: Union[str, bool, None] = None,
                 metadataparent: Optional[str] = None,
                 modelpath: Optional[str] = None,
                 description: Optional[str] = None):
        """
        Initialize a general Parameter object.

        Parameters
        ----------
        name : str
            The name of the parameter.  This should correspond to the name of
            the associated class attribute.
        record : Record
            The Record object that the Value is used with, i.e. which record
            is it contained within.
        recordclass : str or class
            The type of record that this value represents.  Given either as a
            string record style or a Record class.
        defaultvalue : any or None, optional
            The default value to use for the property.  The default value of
            None indicates that there is no default value.
        valuerequired: bool, optional
            Indicates if a value must be given for the property.  If True, then
            checks will be performed that a value is assigned to the property.
        allowedvalues : tuple or None, optional
            A list/tuple of values that the parameter is restricted to have.
            Setting this to None (default) indicates any value is allowed.
        metadatakey: str, bool or None, optional
            The key name to use for the property when constructing the record
            metadata dict.  If set to None (default) then name will be used for
            metadatakey.  If set to False then the parameter will not be
            included in the metadata dict.
        metadataparent: str or None, optional
            If given, then this indicates that the metadatakey is actually an
            element of a dict in metadata with this name.  This allows for limited
            support for metadata having embedded dicts.
        modelpath: str, optional
            The period-delimited path after the record root element for
            where the parameter will be found in the built data model.  If set
            to None (default) then name will be used for modelpath.
        description: str or None, optional
            A short description for the value.  If not given, then the record name
            will be used.
        """
        if issubclass(recordclass, Record):
            self.__recordclass = recordclass
        elif isinstance(recordclass, str):
            self.__recordclass = recordmanager.get_class(recordclass)
        else:
            raise TypeError('invalid recordclass: should be a Record class or string record style')
        self.__emptyrecord = self.recordclass(noname=True)

        super().__init__(name, record, defaultvalue=defaultvalue,
                         valuerequired=valuerequired, allowedvalues=allowedvalues,
                         metadatakey=metadatakey, metadataparent=metadataparent,
                         modelpath=modelpath, description=description)

    @property
    def recordclass(self) -> type[Record]:
        """Class: The record class associated with this value"""
        return self.__recordclass
    
    @property
    def emptyrecord(self) -> Record:
        """Record: An empty record of the appropriate style that is used to check metadata"""
        return self.__emptyrecord
    
    @property
    def _default_queries(self) -> dict:
        """dict: Default query operations to associate with the Value style"""
        if self.metadatakey is False:
            return {}
        elif not hasattr(self, '__default_queries'):
            self.__default_queries = {}
            for key, record_query in self.emptyrecord.queries.items():
                if record_query.parent is not None:
                    # Skip queries that already have parents
                    continue
                
                style = record_query.style
                name = record_query.name
                parent = self.metadatakey
                path = f'{self.record.modelroot}.{self.modelpath}.{".".join(record_query.path.split(".")[1:])}'
                description = record_query.description
                self.__default_queries[key] = load_query(style=style, name=name,
                                                         parent=parent, path=path,
                                                         description=description)

        return self.__default_queries

    def set_value_mod(self, val):
        
        # Set as empty list if None
        if val is None:
            return []

        # Convert val into list
        if isinstance(val, dict):
            val = [val]
        else:
            val = aslist(val)

        for i, v in enumerate(val):
            
            # Check if Record of the correct style, and set noname
            if isinstance(v, Record):
                if v.style != self.emptyrecord.style:
                    raise ValueError('value being set is not of the expected record style')
                v.noname = True
            
            # Build record from dict of init parameters
            elif isinstance(v, dict):
                v['noname'] = True
                val[i] = self.recordclass(**v)
            
            else:
                raise TypeError('value(s) must be Record objects or dicts of Record init parameters')

        return val

    def append(self, **kwargs):
        """
        Appends a Record value to the current list.

        Parameters
        ----------
        val : Record, dict, DM, str, or file-like object
            A record object or the associated content given either as a dict
            of keyword parameters or as a data model.
        """
        kwargs['noname'] = True
        val = self.recordclass(**kwargs)
        self.value.append(val)

    def build_model_value(self):
        models = []
        for val in self.value:
            models.append(val.build_model()[self.emptyrecord.modelroot])

        if len(models) == 0:
            return
        elif len(models) == 1:
            return models[0]
        return models
    
    def metadata_value(self):
        """Function to modify how values are represented in the metadata"""
        metadatas = []
        for val in self.value:
            metadatas.append(val.metadata())
        return metadatas
    
    def load_model_value(self, val):
        if isinstance(val, dict):
            val = [val]
        
        models = []
        for v in val:
            model = DM()
            model[self.emptyrecord.modelroot] = v
            models.append({'model': model})

        return models



# coding: utf-8
# Standard Python libraries
from pathlib import Path
import json
from copy import deepcopy
from typing import Union, Optional

# Relative imports
from yabadaba.tools import screen_input

class Settings():
    """
    Class for handling saved settings.
    """
    def __init__(self,
                 directoryname: str = '.NIST',
                 filename: str = 'settings.json'):
        """
        Class initializer. Calls load.
        
        Parameters
        ----------
        directoryname : str, optional
            The default directory where the settings file is expected will be
            in a directory of this name within the user's home directory.
            Default value is ".NIST".
        filename, str, optional
            The name to use for the settings file.  This will be saved in the
            directory path and should have a ".json" extension.  Default value
            is "settings.json".
        """
        self.__directoryname = directoryname
        self.__filename = filename
        self.__content = {}
        self.__defaultcontent = {}

        self.load()

    ######################## Basic settings operations ########################

    @property
    def defaultdirectory(self) -> Path:
        """pathlib.Path : Path to the default settings directory"""
        return Path(Path.home(), self.__directoryname)

    @property
    def defaultfilename(self) -> Path:
        """pathlib.Path : Path to the default settings file"""
        return Path(self.defaultdirectory, self.__filename)

    @property
    def directory(self) -> Path:
        """pathlib.Path : Path to the settings directory"""
        return self.__directory

    @property
    def filename(self) -> Path:
        """pathlib.Path : Path to the settings file"""
        return Path(self.directory, self.__filename)

    @property
    def content(self) -> dict:
        """dict : The contents of the settings file"""
        return self.__content
    
    @property
    def defaultcontent(self) -> dict:
        """dict : The contents of the default settings file"""
        return self.__defaultcontent
    
    def load(self):
        """Loads the settings file."""
        
        # Load settings.json from the default location
        if self.defaultfilename.is_file():
            with open(self.defaultfilename, 'r') as f:
                self.__defaultcontent = json.load(fp=f)
        else:
            self.__defaultcontent = {}
            
        # Check if forwarding_directory has been set
        if 'forwarding_directory' in self.__defaultcontent:
            self.__directory = Path(self.__defaultcontent['forwarding_directory'])
            
            # Load content from the forwarded location
            if self.filename.is_file():
                with open(self.filename, 'r') as f:
                    self.__content = json.load(fp=f)
            else:
                self.__content = {}

            # Check for recursive forwarding
            if 'forwarding_directory' in self.__content:
                raise ValueError('Multi-level forwarding not allowed.')
        
        # If no forwarding, default is current content
        else:
            self.__content = self.__defaultcontent
            self.__directory = self.defaultdirectory

    def save(self):
        """Saves content to settings.json."""
        if not self.directory.is_dir():
            self.directory.mkdir(parents=True)

        with open(self.filename, 'w') as f:
            json.dump(self.content, fp=f, indent=4)
        
        # Reload
        self.load()

    def set_directory(self,
                      directory: Union[str, Path] = None,
                      move: bool = True,
                      prompt: bool = True):
        """
        Changes the settings directory to a different location.

        Parameters
        ----------
        directory : str or Path
            The path to the new settings directory where settings.json (and the
            default library directory) are to be located.
        move : bool, optional
            If True (default), will attempt to move the exiting settings file to the new
            location.  If False, only the directory path is updated and any settings in the
            old and new locations are left alone.
        prompt : bool, optional
            The default behavior is to prompt for missing details and confirmation of
            changes being made.  Setting prompt = False will skip confirmations and
            throw errors if required inputs are not given to the function.

        Raises
        ------
        TypeError
            If prompt is False and directory is not given.
        ValueError
            If an invalid response is given to the confirmation prompt or if move is True
            and a non-empty settings file already exists in the new location.
        """
        # Manage directory values
        if directory is None:
            if prompt:
                directory = screen_input("Enter the path for the new settings directory:")
            else:
                raise TypeError('directory must be specified if prompt is False')
        directory = Path(directory).resolve()
        
        # Check if given directory path is the current set value
        if self.directory == directory:
            print(f'Settings directory is already set to "{self.directory}"')
            return None
        
        # Prompt confirmation
        if prompt:
            print(f'Current settings directory is "{self.directory}"')
            option = screen_input(f'Update to "{directory}"? (yes or no):')
            if option.lower() in ['yes', 'y']:
                pass
            elif option.lower() in ['no', 'n']: 
                return None
            else: 
                raise ValueError('Invalid choice')
        
        # Move existing settings
        if move:
            
            # Check for file in the new location
            filename = Path(directory, self.__filename)
            if filename.is_file():
                with open(filename, 'r') as f:
                    test_content = f.read()
                if test_content.strip() not in ['', '{}']:
                    raise ValueError(f'cannot move directory as "{filename}" exists and contains content')

            # Move the settings file
            self.filename.replace(filename)
            
            # Clear defaultcontent if previous directory was the default
            if self.directory == self.defaultdirectory:
                self.__defaultcontent = {}
            
        # Update forwarding_directory value in default content
        self.defaultcontent['forwarding_directory'] = directory.as_posix()
        
        # Save the updated default settings file
        if not self.defaultdirectory.is_dir():
            self.defaultdirectory.mkdir(parents=True)
        with open(self.defaultfilename, 'w') as f:
            json.dump(self.defaultcontent, fp=f, indent=4)
        
        # Reload settings
        self.load()
        
    def unset_directory(self,
                        move: bool = True,
                        prompt: bool = True):
        """
        Changes the settings directory back to the default location.
        
        Parameters
        ----------
        move : bool
            If True (default), will attempt to move the contents of the existing settings file
            to the default location.  If False, only the directory path is updated and any settings
            in the old and new locations are left alone.
        prompt : bool, optional
            The default behavior is to prompt for missing details and confirmation of
            changes being made.  Setting prompt = False will skip confirmations and
            throw errors if required inputs are not given to the function.

        Raises
        ------
        ValueError
            If an invalid response is given to the confirmation prompt or if move is True
            and a non-empty settings file already exists in the new location.
        """
        
        # Check if directory already is the default value
        if self.directory == self.defaultdirectory:
            print(f'Settings directory is already set to "{self.defaultdirectory}"')
            return None
        
        # Prompt confirmation
        if prompt:
            print(f'Current settings directory is "{self.directory}"')
            option = screen_input(f'Update to "{self.defaultdirectory}"? (yes or no):')
            if option.lower() in ['yes', 'y']:
                pass
            elif option.lower() in ['no', 'n']: 
                return None
            else: 
                raise ValueError('Invalid choice')
        
        # Move existing settings
        if move:
            # Check if default content contains other settings
            for key in self.defaultcontent:
                if key != 'forwarding_directory':
                    raise ValueError(f'cannot move directory as "{self.defaultfilename}" contains content')

            # Move the settings file to the default location
            self.filename.replace(self.defaultfilename)
        
        # Unset without moving
        else:       
            
            # Remove forwarding_directory pointer
            del self.defaultcontent['forwarding_directory']
            
            # Save the updated default settings file
            if not self.defaultdirectory.is_dir():
                self.defaultdirectory.mkdir(parents=True)
            with open(self.defaultfilename, 'w') as f:
                json.dump(self.defaultcontent, fp=f, indent=4)
        
        # Reload settings
        self.load()
    
    ############################ database settings ############################

    @property
    def databases(self) -> dict:
        """dict: Any defined database settings organized by name"""
        if 'database' in self.content:
            return deepcopy(self.content['database'])
        else:
            return {}

    @property
    def list_databases(self) -> list:
        """list: The names associated with the defined databases"""
        return list(self.databases.keys())

    def set_database(self,
                     name: Optional[str] = None,
                     style: Optional[str] = None,
                     host: Optional[str] = None,
                     prompt: bool = True,
                     **kwargs):
        """
        Allows for database information to be defined in the settings file. Screen
        prompts will be given to allow any necessary database parameters to be
        entered.

        Parameters
        ----------
        name : str, optional
            The name to assign to the database. If not given, the user will be
            prompted to enter one.
        style : str, optional
            The database style associated with the database. If not given, the
            user will be prompted to enter one.
        host : str, optional
            The database host (directory path or url) where the database is
            located. If not given, the user will be prompted to enter one.
        prompt : bool, optional
            The default behavior is to prompt for missing details and confirmation of
            changes being made.  Setting prompt = False will skip confirmations and
            throw errors if required inputs are not given to the function.
        **kwargs : any, optional
            Any other database style-specific parameter settings required to
            properly access the database.

        Raises
        ------
        ValueError
            If the answer to the overwrite question is invalid.
        """
        # Ask for name if not given
        if name is None:
            if prompt:
                name = screen_input('Enter a name for the database:')
            else:
                raise TypeError('name must be specified if prompt is False')

        # Load database if it exists
        if name in self.list_databases:
            
           # Ask if existing database should be overwritten
            if prompt:
                print(f'Database {name} already defined.')
                option = screen_input('Overwrite? (yes or no):')
                if option.lower() in ['yes', 'y']:
                    pass
                elif option.lower() in ['no', 'n']: 
                    return None
                else: 
                    raise ValueError('Invalid choice')

        # Ask for style if not given
        if style is None: 
            if prompt:
                style = screen_input("Enter the database's style:")
            else:
                raise TypeError('style must be specified if prompt is False')
        
        # Ask for host if not given
        if host is None: 
            if prompt:
                host = screen_input("Enter the database's host:")
            else:
                raise TypeError('host must be specified if prompt is False')
        
        if len(kwargs) == 0 and prompt:
            print('Enter any other database parameters as key, value')
            print('Exit by leaving key blank')
            while True:
                key = screen_input('key:')
                if key == '': 
                    break
                kwargs[key] = screen_input('value:')

        # Create database entry
        if 'database' not in self.content:
            self.content['database'] = {}
        self.content['database'][name] = entry = {}
        entry['style'] = style
        entry['host'] = str(host)
        for key, value in kwargs.items():
            entry[key] = value

        # Save changes
        self.save()
    
    def unset_database(self,
                       name: Optional[str] = None,
                       prompt: bool = True):
        """
        Deletes the settings for a pre-defined database from the settings file.

        Parameters
        ----------
        name : str, optional
            The name assigned to a pre-defined database.
        prompt : bool, optional
            The default behavior is to prompt for missing details and confirmation of
            changes being made.  Setting prompt = False will skip confirmations and
            throw errors if required inputs are not given to the function.

        Raises
        ------
        TypeError
            If prompt is False and name is not given.
        ValueError
            If the database name is not found
        """
        database_names = self.list_databases
                  
        # Ask for name if not given
        if name is None:
            if len(database_names) > 0:
                if prompt:
                    print('Select a database:')
                    for i, database in enumerate(database_names):
                        print(i+1, database)
                    choice = screen_input(':')
                    try:
                        choice = int(choice)
                    except:
                        name = choice
                    else:
                        name = database_names[choice-1]
                else:
                    raise TypeError('name must be specified if prompt is False')
            else:
                print('No databases currently set')
                return None

        # Verify listed name exists 
        try:
            i = database_names.index(name)
        except:
            raise ValueError(f'Database {name} not found')

        # Confirmation prompt
        if prompt:
            print(f'Database {name} found')
            test = screen_input('Delete settings? (must type yes):').lower()
            if test != 'yes':
                return None

        # Delete database and save
        del(self.content['database'][name])
        if len(self.content['database']) == 0:
            del(self.content['database'])
        self.save()

# Initialize settings
settings = Settings()


@property 
def degFinterval(self) -> float:
    """temperature difference in degrees Fahrenheit"""
    return (5./9.) * self.K

@property
def degCinterval(self) -> float:
    """temperature difference in degrees Celsius"""
    return self.K

@property
def mK(self) -> float:
    """millikelvin"""
    return 1e-3 * self.K

@property
def uK(self) -> float:
    """microkelvin"""
    return 1e-6 * self.K

@property
def nK(self) -> float:
    """nanokelvin"""
    return 1e-9 * self.K

@property
def pK(self) -> float:
    """picokelvin"""
    return 1e-12 * self.K


from typing import Optional, Union, Any

from DataModelDict import DataModelDict as DM

import numpy as np

from ..query import load_query
from . import Value

class BoolValue(Value):
    
    def set_value_mod(self, val):
        
        # Check if value is in #text
        val = self.set_value_mod_textfield(val)
        
        # Pass Boolean values through without changing
        if val is True or val is False or val is None:
            return val

        if isinstance(val, str):
            # Convert strings
            if val.lower() in ['true', 't']:
                return True
            elif val.lower() in ['false', 'f']:
                return False
            else:
                raise ValueError(f'String value for {self.name} not recognized as a Boolean value.')
        else:
            raise TypeError(f'Invalid type for Boolean {self.name}. Must be True, False, None, or a str representation of True or False')
    
    @property
    def _default_queries(self) -> dict:
        """dict: Default query operations to associate with the Parameter style"""
        if self.metadatakey is False:
            return {}
        else:
            return {
                self.name: load_query('bool_match',
                                    name=self.metadatakey,
                                    parent=self.metadataparent,
                                    path=f'{self.record.modelroot}.{self.modelpath}',
                                    description=f'Return only the records where {self.description} matches a given value')
            }


from typing import Optional, Union, Any
from copy import deepcopy

from DataModelDict import DataModelDict as DM

from ..query import load_query
from ..tools import dict_insert

class Value():

    def __init__(self,
                 name: str,
                 record,
                 defaultvalue: Optional[Any] = None,
                 valuerequired: bool = False,
                 allowedvalues: Optional[tuple] = None,
                 metadatakey: Union[str, bool, None] = None,
                 metadataparent: Optional[str] = None,
                 modelpath: Optional[str] = None,
                 description: Optional[str] = None):
        """
        Initialize a Value object.

        Parameters
        ----------
        name : str
            The name of the parameter.  This should correspond to the name of
            the associated class attribute.
        record : Record, optional
            The Record object that the Parameter is used with.
        defaultvalue : any or None, optional
            The default value to use for the property.  The default value of
            None indicates that there is no default value.
        valuerequired: bool, optional
            Indicates if a value must be given for the property.  If True, then
            checks will be performed that a value is assigned to the property.
        allowedvalues : tuple or None, optional
            A list/tuple of values that the parameter is restricted to have.
            Setting this to None (default) indicates any value is allowed.
        metadatakey: str, bool or None, optional
            The key name to use for the property when constructing the record
            metadata dict.  If set to None (default) then name will be used for
            metadatakey.  If set to False then the parameter will not be
            included in the metadata dict.
        metadataparent: str or None, optional
            If given, then this indicates that the metadatakey is actually an
            element of a dict in metadata with this name.  This allows for limited
            support for metadata having embedded dicts.
        modelpath: str, optional
            The period-delimited path after the record root element for
            where the parameter will be found in the built data model.  If set
            to None (default) then name will be used for modelpath.
        description: str or None, optional
            A short description for the value.  If not given, then the record name
            will be used.
        """

        self.__name = str(name)
        self.__record = record
        self.__defaultvalue = defaultvalue
        self.__valuerequired = valuerequired
        self.__allowedvalues = allowedvalues

        if description is None:
            description = self.name
        self.__description = description

        if metadatakey is None:
            metadatakey = self.name
        elif isinstance(metadatakey, bool):
            if metadatakey is True:
                raise TypeError('metadatakey can be None, False, or str')
        else:
            metadatakey = str(metadatakey)   
        self.__metadatakey = metadatakey

        self.__metadataparent = metadataparent

        if modelpath is None:
            modelpath = name
        self.__modelpath = str(modelpath)

        self.value = defaultvalue

        self.__queries = deepcopy(self._default_queries)


    @property
    def name(self) -> str:
        """str: The name of the parameter"""
        return self.__name

    @property
    def record(self):
        """Record: The record that this parameter is used with"""
        return self.__record

    @property
    def defaultvalue(self) -> Any:
        """any: The default value to use for the parameter"""
        return self.__defaultvalue
    
    @property
    def valuerequired(self) -> bool:
        """bool: Indicates if the parameter must be given"""
        return self.__valuerequired
    
    @property
    def metadatakey(self) -> Union[str, bool]:
        """str or bool: The key to save the property under in the metadata dict"""
        return self.__metadatakey
    
    @property
    def metadataparent(self) -> Optional[str]:
        """str or None: The parent key in the metadata dict that metadatakey is below"""
        return self.__metadataparent
    
    @property
    def modelpath(self) -> str:
        """str: The period-delimited path after the root element where the parameter is stored in a data model"""
        return self.__modelpath
    
    @property
    def description(self) -> str:
        """str: A short description of the value"""
        return self.__description
    
    @description.setter
    def description(self, val: str):
        self.__description = str(val)

    @property
    def queries(self) -> dict:
        """dict: The Query operations associated with this parameter"""
        return self.__queries
    
    @property
    def value(self) -> Any:
        """any: The value assigned to the parameter"""
        if self.__value is None and self.valuerequired:
            raise ValueError(f'{self.name} has not been set!')
        return self.__value
    
    @value.setter
    def value(self, val):
        val = self.set_value_mod(val)
        
        if self.allowedvalues is not None and val is not None and val not in self.allowedvalues:
            raise ValueError(f'Invalid value {val} for {self.name}. Allowed values are {self.allowedvalues}')
            
        self.__value = val

    def set_value_mod(self, val):
        """Modifies values before setting them"""            
        return val

    @staticmethod
    def set_value_mod_textfield(val: Any) -> Any:
        """
        Utility method for use in set_value_mod that checks for the case where
        a simple value is being set from its modelpath and that element may or
        may not contain XML attributes. 
        """
        # Mod for reading model elements with attributes
        if isinstance(val, dict):
            if '#text' in val:
                val = val['#text']
            else:
                return None
            
        return val

    @property
    def allowedvalues(self) -> Optional[tuple]:
        """tuple or None: tuple containing all allowed values.  If None, no limits."""
        return self.__allowedvalues
    
    @allowedvalues.setter
    def allowedvalues(self, val: Optional[tuple]):
        if val is None:
            self.__allowedvalues = None
        else:
            self.__allowedvalues = tuple(val)

    @property
    def _default_queries(self) -> dict:
        """dict: Default query operations to associate with the Value style"""
        return {}

    def add_query(self,
                  style: str,
                  description: str,
                  name: Optional[str] = None,
                  metadatakey: Optional[str] = None,
                  metadataparent: Optional[str] = None,
                  modelpath: Optional[str] = None):
        """
        Builds and adds a new query operation for the parameter using the
        parameter's settings.

        Parameters
        ----------
        style : str
            The style of Query to add.
        description : str
            The doc description of the Query operation.
        name : str or None, optional
            The name of the keyword argument that will be used in the query
            functions for this Query operation.  If None (default) then the
            name of the parameter will be used which is ideal if only one
            query is associated with the parameter.
        metadatakey : str or None, optional
            The element key of the metadata where the value is.  If None
            (default) then the corresponding parameter value is used.
        metadataparent : str or None, optional
            Allows for the pandas query operations to work on embedded
            metadata dicts.  If given, the pandas query method will check the
            value of metadata[parent][name].
        modelpath : str or None, optional
            The period-delimited path after the record's modelroot where the
            value is found in the model. If None (default) then the corresponding
            parameter value is used.
        """
        if name is None:
            name = self.name
        if metadatakey is None and self.metadatakey is not False:
            metadatakey = self.metadatakey
            metadataparent = self.metadataparent
        if modelpath is None:
            modelpath = self.modelpath
        
        self.queries[name] = load_query(style, name=metadatakey,
                                        parent=metadataparent,
                                        path=f'{self.record.modelroot}.{modelpath}',
                                        description=description)

    def build_model(self,
                    model,
                    **kwargs):
        """
        Inserts the parameter model into the record's model.
        
        Parameters
        ----------
        model : DataModelDict.DataModelDict
            The record content (after root element) to add content to.
        before : any, optional
            An element key at the same final level as this parameter's model
            path that this parameter should be inserted directly before.
            Cannot be given with after.  If neither before nor after is given
            then the parameter will be inserted into the dict normally.
        after : any, optional
            An element key at the same final level as this parameter's model
            path that this parameter should be inserted directly after.
            Cannot be given with before.  If neither before nor after is given
            then the parameter will be inserted into the dict normally.
        """
        # Build value to output
        value = self.build_model_value()
        
        # Do nothing if value not required and empty
        if value is None and self.valuerequired is False:
            return

        # Split path
        path = self.modelpath.split('.')
        
        # Iterate down path and create parent elements if needed
        m = model
        for key in path[:-1]:
            if key not in m:
                m[key] = DM()
            m = m[key]
        
        # Check if target path exists and is a dict
        if path[-1] in m and isinstance(m[path[-1]], dict):
            # Insert the value as #text
            dict_insert(m[path[-1]], '#text', value, **kwargs)

        else:
            # Insert the parameter's value into the model
            dict_insert(m, path[-1], value, **kwargs)


    def build_model_value(self):
        """Function to modify how values are represented in the model"""
        return self.value
    
    def load_model(self, model, setvalue=True) -> Optional[Any]:
        """
        Loads the parameter value from the record's model.
        
        Parameters
        ----------
        model : DataModelDict.DataModelDict
            The record content (after root element) to read.
        setvalue : bool, optional
            If True (default) then the parameter's value will automatically
            be set to the value found at modelpath.  If False, then the
            value at modelpath will be returned instead for further
            manipulations.
        """
        path = self.modelpath.split('.')

        try:
            val = self.load_model_value(model[path])
        except (KeyError, TypeError):
            val = self.defaultvalue

        if setvalue is True:
            self.value = val
        else:
            return val

    def load_model_value(self, val):
        """Function to modify how values are interpreted from the model"""
        return val

    def metadata(self, meta):
        """
        Adds the parameter to the record's metadata dict.

        Parameters
        ----------
        meta : dict
            The metadata dict being built for the record.
        """

        if self.metadatakey is False:
            return
        
        if self.metadataparent is None:
            meta[self.metadatakey] = self.metadata_value()

        else:
            if self.metadataparent not in meta:
                meta[self.metadataparent] = {}
            meta[self.metadataparent][self.metadatakey] = self.metadata_value()

    def metadata_value(self):
        """Function to modify how values are represented in the metadata"""
        return self.value


import math

@property
def NA(self) -> float:
    """Avogadro's number"""
    return 6.02214076e23

@property
def pi(self) -> float:
    """pi"""
    return math.pi

@property
def π(self) -> float:
    """pi"""
    return math.pi

@property
def c0(self) -> float:
    """speed of light in vacuum"""
    return 299792458. * self.m / self.s

@property
def mu0(self) -> float:
    """magnetic constant, permeability of vacuum"""
    return 1.25663706212e-6 * self.N / self.A**2

@property
def μ0(self) -> float:
    """magnetic constant, permeability of vacuum"""
    return self.mu0

@property
def eps0(self) -> float:
    """electric constant, permittivity of vacuum"""
    return 1. / (self.mu0 * self.c0**2)

@property
def ε0(self) -> float:
    """electric constant, permittivity of vacuum"""
    return self.eps0

@property
def Z0(self) -> float:
    """vacuum impedance, 377 ohms"""
    return self.mu0 * self.c0

@property
def hPlanck(self) -> float:
    """planck constant"""
    return 6.62607015e-34 * self.J * self.s

@property
def hbar(self) -> float:
    """reduced planck constant"""
    return self.hPlanck / (2. * self.pi)

@property
def ħ(self) -> float:
    """reduced planck constant"""
    return self.hbar

@property
def kB(self) -> float:
    """Boltzmann constant"""
    return 1.380649e-23 * self.J / self.K

@property
def GNewton(self) -> float:
    """Gravitational constant"""
    return 6.67430e-11 * self.m**3 / (self.kg * self.s**2)

@property
def sigmaSB(self) -> float:
    """Stefan-Boltzmann constant"""
    return (self.pi**2 / 60.) * self.kB**4 / (self.hbar**3 * self.c0**2)

@property
def σSB(self) -> float:
    """Stefan-Boltzmann constant"""
    return self.sigmaSB

@property
def alphaFS(self) -> float:
    """fine-structure constant"""
    return 7.2973525693e-3

@property
def αFS(self) -> float:
    """fine-structure constant"""
    return self.alphaFS

############# Constants--chemistry, atomic physics, electrons #############

@property
def Rgas(self) -> float:
    """ideal gas constant (see README)"""
    return self.kB

@property
def e(self) -> float:
    """charge of proton"""
    return 1.602176634e-19 * self.C 

@property
def uBohr(self) -> float:
    """Bohr magneton"""
    return 9.2740100783e-24 * self.J / self.T

@property
def uNuc(self) -> float:
    """nuclear magneton"""
    return 5.0507837461e-27 * self.J / self.T

@property
def aBohr(self) -> float:
    """Bohr radius"""
    return 5.29177210903e-11 * self.m

@property
def me(self) -> float:
    """electron mass"""
    return 9.1093837015e-31 * self.kg

@property
def mp(self) -> float:
    """proton mass"""
    return 1.67262192369e-27 * self.kg

@property
def mn(self) -> float:
    """neutron mass"""
    return 1.67492749804e-27 * self.kg

@property
def Rinf(self) -> float:
    """Rydberg constant"""
    return 10973731.568160 / self.m

@property
def Ry(self) -> float:
    """Rydberg energy, approximately 13.6 eV"""
    return 2.1798723611035e-18 * self.J

@property
def Hartree(self) -> float:
    """Hartree energy, approximately 27.2 eV"""
    return 2 * self.Ry

@property
def ARichardson(self) -> float:
    """Richardson constant"""
    return (4. * self.pi * self.e * self.me * self.kB**2) / self.hPlanck**3

@property
def Phi0(self) -> float:
    """magnetic flux quantum"""
    return self.hPlanck / (2 * self.e)

@property
def KJos(self) -> float:
    """Josephson constant"""
    return (2 * self.e) / self.hPlanck

@property
def RKlitz(self) -> float:
    """von Klitzing constant"""
    return self.hPlanck / self.e**2

@property
def debye(self) -> float:
    """debye dipole moment, approximately 0.0208 e nm"""
    return 1e-21 * self.C * self.m**2 / (self.s * self.c0)

############## Constants--astronomical and properties of earth ##############

@property
def REarth(self) -> float:
    """radius of earth"""
    return 6371. * self.km

@property
def g0(self) -> float:
    """standard earth gravitational acceleration"""
    return 9.80665 * self.m / self.s**2

@property
def Msolar(self) -> float:
    """mass of the sun)"""
    return 1.98847e30 * self.kg

@property
def MEarth(self) -> float:
    """mass of earth"""
    return 5.9722e24 * self.kg